"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.outdatedRecursive = void 0;
const cli_utils_1 = require("@pnpm/cli-utils");
const error_1 = require("@pnpm/error");
const outdated_1 = require("@pnpm/outdated");
const table_1 = require("@zkochan/table");
const chalk_1 = __importDefault(require("chalk"));
const isEmpty_1 = __importDefault(require("ramda/src/isEmpty"));
const sortWith_1 = __importDefault(require("ramda/src/sortWith"));
const outdated_2 = require("./outdated");
const utils_1 = require("./utils");
const DEP_PRIORITY = {
    dependencies: 1,
    devDependencies: 2,
    optionalDependencies: 0,
};
const COMPARATORS = [
    ...utils_1.DEFAULT_COMPARATORS,
    (o1, o2) => DEP_PRIORITY[o1.belongsTo] - DEP_PRIORITY[o2.belongsTo],
];
async function outdatedRecursive(pkgs, params, opts) {
    const outdatedMap = {};
    const rootManifest = pkgs.find(({ dir }) => dir === opts.lockfileDir ?? opts.dir);
    const outdatedPackagesByProject = await (0, outdated_1.outdatedDepsOfProjects)(pkgs, params, {
        ...opts,
        fullMetadata: opts.long,
        ignoreDependencies: new Set(rootManifest?.manifest?.pnpm?.updateConfig?.ignoreDependencies ?? []),
        retry: {
            factor: opts.fetchRetryFactor,
            maxTimeout: opts.fetchRetryMaxtimeout,
            minTimeout: opts.fetchRetryMintimeout,
            retries: opts.fetchRetries,
        },
        timeout: opts.fetchTimeout,
    });
    for (let i = 0; i < outdatedPackagesByProject.length; i++) {
        const { dir, manifest } = pkgs[i];
        outdatedPackagesByProject[i].forEach((outdatedPkg) => {
            const key = JSON.stringify([outdatedPkg.packageName, outdatedPkg.current, outdatedPkg.belongsTo]);
            if (!outdatedMap[key]) {
                outdatedMap[key] = { ...outdatedPkg, dependentPkgs: [] };
            }
            outdatedMap[key].dependentPkgs.push({ location: dir, manifest });
        });
    }
    let output;
    switch (opts.format ?? 'table') {
        case 'table': {
            output = renderOutdatedTable(outdatedMap, opts);
            break;
        }
        case 'list': {
            output = renderOutdatedList(outdatedMap, opts);
            break;
        }
        case 'json': {
            output = renderOutdatedJSON(outdatedMap, opts);
            break;
        }
        default: {
            throw new error_1.PnpmError('BAD_OUTDATED_FORMAT', `Unsupported format: ${opts.format?.toString() ?? 'undefined'}`);
        }
    }
    return {
        output,
        exitCode: (0, isEmpty_1.default)(outdatedMap) ? 0 : 1,
    };
}
exports.outdatedRecursive = outdatedRecursive;
function renderOutdatedTable(outdatedMap, opts) {
    if ((0, isEmpty_1.default)(outdatedMap))
        return '';
    const columnNames = [
        'Package',
        'Current',
        'Latest',
        'Dependents',
    ];
    const columnFns = [
        outdated_2.renderPackageName,
        outdated_2.renderCurrent,
        outdated_2.renderLatest,
        dependentPackages,
    ];
    if (opts.long) {
        columnNames.push('Details');
        columnFns.push(outdated_2.renderDetails);
    }
    // Avoid the overhead of allocating a new array caused by calling `array.map()`
    for (let i = 0; i < columnNames.length; i++)
        columnNames[i] = chalk_1.default.blueBright(columnNames[i]);
    const data = [
        columnNames,
        ...sortOutdatedPackages(Object.values(outdatedMap))
            .map((outdatedPkg) => columnFns.map((fn) => fn(outdatedPkg))),
    ];
    return (0, table_1.table)(data, {
        ...cli_utils_1.TABLE_OPTIONS,
        columns: {
            ...cli_utils_1.TABLE_OPTIONS.columns,
            // Dependents column:
            3: {
                width: (0, outdated_2.getCellWidth)(data, 3, 30),
                wrapWord: true,
            },
        },
    });
}
function renderOutdatedList(outdatedMap, opts) {
    if ((0, isEmpty_1.default)(outdatedMap))
        return '';
    return sortOutdatedPackages(Object.values(outdatedMap))
        .map((outdatedPkg) => {
        let info = `${chalk_1.default.bold((0, outdated_2.renderPackageName)(outdatedPkg))}
${(0, outdated_2.renderCurrent)(outdatedPkg)} ${chalk_1.default.grey('=>')} ${(0, outdated_2.renderLatest)(outdatedPkg)}`;
        const dependents = dependentPackages(outdatedPkg);
        if (dependents) {
            info += `\n${chalk_1.default.bold(outdatedPkg.dependentPkgs.length > 1
                ? 'Dependents:'
                : 'Dependent:')} ${dependents}`;
        }
        if (opts.long) {
            const details = (0, outdated_2.renderDetails)(outdatedPkg);
            if (details) {
                info += `\n${details}`;
            }
        }
        return info;
    })
        .join('\n\n') + '\n';
}
function renderOutdatedJSON(outdatedMap, opts) {
    const outdatedPackagesJSON = sortOutdatedPackages(Object.values(outdatedMap))
        .reduce((acc, outdatedPkg) => {
        acc[outdatedPkg.packageName] = {
            current: outdatedPkg.current,
            latest: outdatedPkg.latestManifest?.version,
            wanted: outdatedPkg.wanted,
            isDeprecated: Boolean(outdatedPkg.latestManifest?.deprecated),
            dependencyType: outdatedPkg.belongsTo,
            dependentPackages: outdatedPkg.dependentPkgs.map(({ manifest, location }) => ({ name: manifest.name, location })),
        };
        if (opts.long) {
            acc[outdatedPkg.packageName].latestManifest = outdatedPkg.latestManifest;
        }
        return acc;
    }, {});
    return JSON.stringify(outdatedPackagesJSON, null, 2);
}
function dependentPackages({ dependentPkgs }) {
    return dependentPkgs
        .map(({ manifest, location }) => manifest.name ?? location)
        .sort()
        .join(', ');
}
function sortOutdatedPackages(outdatedPackages) {
    return (0, sortWith_1.default)(COMPARATORS, outdatedPackages.map(outdated_2.toOutdatedWithVersionDiff));
}
//# sourceMappingURL=recursive.js.map